// C Source File
// Created 15/08/2004; 17:11:07

#include <tigcclib.h>

#include "..\Structs.h"
#include "..\Tables\Tbl_Tags.h"
#include "..\Tables\Tbl_Tokens.h"
#include "..\Tables\Tbl_Lst.h"
#include "..\System\String.h"
#include "..\System\File.h"
#include "..\System\Error.h"
#include "..\Other\Scanner.h"
#include "..\Other\Tokenisor.h"
#include "..\Other\ExprVar.h"
#include "..\Other\ExprTokens.h"


//********************************
//Cration du buffer d'execution
//********************************
BOOL ExprTokens_MakeBuffer (void)
{
	unsigned char type, tag, subtag, token_type, token_index, symbol_index;
	
	//Parcours la pile
	for (int i=0; i<TStackExpr_Ptr; i++)
	{
		type = TStackExpr [i]->type; tag = TStackExpr [i]->tag; subtag = TStackExpr [i]->subtag; 
		token_type = TStackExpr [i]->token_type; token_index = TStackExpr [i]->token_index;
		symbol_index = TStackExpr [i]->symbol_index;
		
		//ENDTAG
		if (type==TYPE_ENDTAG) ExprTokens_Add (NULL, ETOKEN_ENDTAG, 1);
		//Nombre
		else if (tag==TAG_NUMBER)
		{
			if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
			ExprTokens_Add (NULL, (subtag==SUBTAG_FLOAT)?ETOKEN_NUMBER_FLOAT:ETOKEN_NUMBER_INT, 1);	//Token
			
			//Float
			if (subtag==SUBTAG_FLOAT)	ExprTokens_Add (&TStackExpr [i]->fnum, 0, sizeof (float));
			//Entier
			else ExprTokens_Add (&TStackExpr [i]->lnum, 0, sizeof (long));
		}
		//Chane de caractre
		else if (tag==TAG_STRING)
		{
			if (TokensFile_size&1) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
			ExprTokens_Add (NULL, ETOKEN_STRING, 1);	//Token
			ExprTokens_Add (NULL, TStackExpr [i]->string.size, 1);	//Taille de la chane
			ExprTokens_Add (TStackExpr [i]->string.data, 0, TStackExpr [i]->string.size);	//Adresse de la chane
			ExprTokens_Add (NULL, 0, 1);	//Octet NULL
		}
		//Variable
		else if (tag==TAG_VAR)
		{
			ExprTokens_Add (NULL, ETOKEN_VAR_FLOAT+(subtag-1), 1);	//Token
			ExprTokens_Add (NULL, symbol_index, 1);	//Index
		}
		//Tableaux
		else if (tag==TAG_VARTABLE)
		{
			ExprTokens_Add (NULL, ETOKEN_VAR_TFLOAT+(subtag-1), 1);	//Token
			ExprTokens_Add (NULL, symbol_index, 1);	//Index
		}
		//Label
		else if (tag==TAG_LABEL)
			TblLabelAddress [symbol_index] = TokensFile_size;	//Adresse du label
		//Goto
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_GOTO)
		{
			if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
			ExprTokens_Add (NULL, ETOKEN_JUMP, 1);	//Token
			
			//Adresse connue
			if (TblLabelAddress [symbol_index]) ExprTokens_Add (&TblLabelAddress [symbol_index], 0, 2); //Adresse du label
			else
			{
				if (nTblGotoJump>=MAX_GOTO) {error_code = ERROR_GOTO_OVERFLOW; return FALSE;}
				TblGotoJump [nTblGotoJump].lbl_index = symbol_index;
				TblGotoJump [nTblGotoJump++].token_address = TokensFile_size;
				
				//Adresse
				ExprTokens_Add (NULL, ETOKEN_NONE, 1);
				ExprTokens_Add (NULL, ETOKEN_NONE, 1);
			}
		}
		//If
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_IF)
		{
			if (n_IF_Jump>=MAX_IF) {error_code = ERROR_IF_OVERFLOW; return FALSE;}
			if (!ExprTokens_NewStruct (STRUCT_IF)) return FALSE; //Ajoute structure
			
			if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
			ExprTokens_Add (NULL, ETOKEN_IF, 1);	//Token
			
			//Token_address
			Tbl_IF_Jump [n_IF_Jump].nElseIf = 0;
			Tbl_IF_Jump [n_IF_Jump].branch_else = FALSE;
			Tbl_IF_Jump [n_IF_Jump++].token_address = TokensFile_size;
			
			//Adresse
			ExprTokens_Add (NULL, ETOKEN_NONE, 1);
			ExprTokens_Add (NULL, ETOKEN_NONE, 1);
		}
		//Elseif
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_ELSEIF)
		{
			if (!n_IF_Jump || Tbl_IF_Jump [n_IF_Jump-1].branch_else || !nTblStruct || TblStruct [nTblStruct-1]!=STRUCT_IF) 
				{error_code = ERROR_ELSEIF; return FALSE;}
			
			//Jump
			//====
			if (i!=TStackExpr_Ptr-1)
			{	
				if (TotalElseIf>=MAX_ELSEIF) {error_code = ERROR_ELSEIF_OVERFLOW; return FALSE;}
				
				//Ecriture de Jump
				if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
				ExprTokens_Add (NULL, ETOKEN_JUMP, 1);	//Token
				
				//Branchement aprs Jump
				*(unsigned short *)(TokensFile_buffer+Tbl_IF_Jump [n_IF_Jump-1].token_address) = TokensFile_size+2;
				Tbl_IF_Jump [n_IF_Jump-1].nElseIf++;
				Tbl_ElseIf_Address [TotalElseIf++] = TokensFile_size;
				
				//Adresse
				ExprTokens_Add (NULL, ETOKEN_NONE, 1);
				ExprTokens_Add (NULL, ETOKEN_NONE, 1);
				
				//Rinitialisation des piles d'execution
				if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
				ExprTokens_Add (NULL, ETOKEN_CLEAR, 1);
			}
			//Condition
			else
			{
				if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
				ExprTokens_Add (NULL, ETOKEN_IF, 1);	//Token
			
				//Token_address
				Tbl_IF_Jump [n_IF_Jump-1].token_address = TokensFile_size;
			
				//Adresse
				ExprTokens_Add (NULL, ETOKEN_NONE, 1);
				ExprTokens_Add (NULL, ETOKEN_NONE, 1);
			}
		}
		//Else
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_ELSE)
		{
			if (!n_IF_Jump || Tbl_IF_Jump [n_IF_Jump-1].branch_else || !nTblStruct || TblStruct [nTblStruct-1]!=STRUCT_IF) 
				{error_code = ERROR_ELSE; return FALSE;}
			
			//Ecriture de Jump
			if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
			ExprTokens_Add (NULL, ETOKEN_JUMP, 1);	//Token
			
			//Branchement aprs Jump
			*(unsigned short *)(TokensFile_buffer+Tbl_IF_Jump [n_IF_Jump-1].token_address) = TokensFile_size+2;
			Tbl_IF_Jump [n_IF_Jump-1].branch_else = TRUE;
			Tbl_IF_Jump [n_IF_Jump-1].token_address = TokensFile_size;
			
			//Adresse
			ExprTokens_Add (NULL, ETOKEN_NONE, 1);
			ExprTokens_Add (NULL, ETOKEN_NONE, 1);
		}
		//Endif
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_ENDIF)
		{
			if (!n_IF_Jump || !nTblStruct || TblStruct [nTblStruct-1]!=STRUCT_IF) 
				{error_code = ERROR_ENDIF; return FALSE;}
			
			unsigned char nElseif = Tbl_IF_Jump [n_IF_Jump-1].nElseIf;
			for (unsigned short i=TotalElseIf-nElseif; i<TotalElseIf; i++)
				*(unsigned short *)(TokensFile_buffer+Tbl_ElseIf_Address [i]) = TokensFile_size;
			*(unsigned short *)(TokensFile_buffer+Tbl_IF_Jump [--n_IF_Jump].token_address) = TokensFile_size;
			TotalElseIf -= nElseif;
			
			//Efface structure
			ExprTokens_DelStruct (0);
		}
		//Do
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_DO)
		{
			if (TotalDo>=MAX_DO) {error_code = ERROR_DO_OVERFLOW; return FALSE;}
			if (!ExprTokens_NewStruct (STRUCT_DO)) return FALSE; //Ajoute structure
			
			Tbl_Do_Address [TotalDo++] = TokensFile_size;
		}
		//Loop
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_LOOP)
		{
			if (!TotalDo || !nTblStruct || TblStruct [nTblStruct-1]!=STRUCT_DO) 
				{error_code = ERROR_LOOP; return FALSE;}
			
			//Ecriture de Jump
			if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
			ExprTokens_Add (NULL, ETOKEN_JUMP, 1);	//Token
			
			//Adresse
			ExprTokens_Add (&Tbl_Do_Address [--TotalDo], 0, sizeof (unsigned short));
			
			//Efface structure
			ExprTokens_DelStruct (TokensFile_size); 
		}
		//Repeat
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_REPEAT)
		{
			if (TotalRepeat>=MAX_REPEAT) {error_code = ERROR_REPEAT_OVERFLOW; return FALSE;}
			if (!ExprTokens_NewStruct (STRUCT_REPEAT)) return FALSE; //Ajoute structure
			
			Tbl_Repeat_Address [TotalRepeat++] = TokensFile_size;
		}
		//Until
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_UNTIL)
		{
			if (!TotalRepeat || !nTblStruct || TblStruct [nTblStruct-1]!=STRUCT_REPEAT) 
				{error_code = ERROR_UNTIL; return FALSE;}
				
			//Ecriture de If
			if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
			ExprTokens_Add (NULL, ETOKEN_IF, 1);	//Token
			
			//Adresse
			ExprTokens_Add (&Tbl_Repeat_Address [--TotalRepeat], 0, sizeof (unsigned short));
			
			//Efface structure
			ExprTokens_DelStruct (TokensFile_size);	
		}
		//While
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_WHILE)
		{
			//While Jump
			if (i!=TStackExpr_Ptr-1)
			{
				if (TotalWhile>=MAX_WHILE) {error_code = ERROR_WHILE_OVERFLOW; return FALSE;}
				if (!ExprTokens_NewStruct (STRUCT_WHILE)) return FALSE; //Ajoute structure
				
				Tbl_While_Address_Jump [TotalWhile] = TokensFile_size;
			}
			//While If
			else
			{
				//Ecriture de If
				if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
			
				ExprTokens_Add (NULL, ETOKEN_IF, 1);	//Token
				
				Tbl_While_Address_If [TotalWhile++] = TokensFile_size;
				ExprTokens_Add (NULL, ETOKEN_NONE, 1);
				ExprTokens_Add (NULL, ETOKEN_NONE, 1);
			}
		}
		//Wend
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_WEND)
		{
			if (!TotalWhile || !nTblStruct || TblStruct [nTblStruct-1]!=STRUCT_WHILE) 
				{error_code = ERROR_WEND; return FALSE;}
			
			//Saut pour dbut de While
			if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
			ExprTokens_Add (NULL, ETOKEN_JUMP, 1);	//Token
			ExprTokens_Add (&Tbl_While_Address_Jump [--TotalWhile], 0, sizeof (unsigned short));	//Adresse
			
			//Saut pour la condition
			*(unsigned short *)(TokensFile_buffer+Tbl_While_Address_If [TotalWhile]) = TokensFile_size;
			
			//Efface structure
			ExprTokens_DelStruct (TokensFile_size);
		}
		
		//Exit If
		else if (token_type==TOKEN_COMMAND && token_index==TOKEN_EXIT)
		{
			if (!nTblStructExitIf) {error_code = ERROR_EXITIF; return FALSE;}
			if (TotalExitIf>=MAX_EXITIF) {error_code = ERROR_EXITIF_OVERFLOW; return FALSE;}
			
			if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
			
			//Exit seule
			if (TStackExpr_Ptr==1) ExprTokens_Add (NULL, ETOKEN_JUMP, 1);	//Token
			//Exit If
			else ExprTokens_Add (NULL, ETOKEN_IFJUMP, 1);	//Token			
			
			nTblExitIf [nTblStructExitIf-1]++;
			TblExitIf [TotalExitIf++] = TokensFile_size;
			
			//Adresse
			ExprTokens_Add (NULL, ETOKEN_NONE, 1);
			ExprTokens_Add (NULL, ETOKEN_NONE, 1); 
		}
		
		//Autre
		else if (tag==TAG_POINT_VIRGULE) ExprTokens_Add (NULL, ETOKEN_POINT_VIRGULE, 1);
		else if (tag==TAG_APOST) ExprTokens_Add (NULL, ETOKEN_APOST, 1);
		//Fonctions
		else ExprTokens_Add (NULL, (token_type==TOKEN_COMMAND)?GFA_TableExecToken_Command [token_index]:GFA_TableExecToken_Function [token_index], 1);
	}
	
	//N'crit pas une ligne vierge ou une ligne de commantaire
	if (TStackExpr_Ptr)
	{
		if (!(TokensFile_size&1)) ExprTokens_Add (NULL, ETOKEN_NONE, 1);	//Alignement
		ExprTokens_Add (NULL, ETOKEN_EOFLINE, 1);
		ExprTokens_Add (&error_line, 0, 2);
	}
	
	return TRUE;
}


//***************************************
//Attribution des adresses pour les GOTOS
//***************************************
void ExprTokens_MakeGotoJump (void)
{
	unsigned short address, token_address;
	
	//Parcours sauts  effectuer
	for (unsigned short i=0; i<nTblGotoJump; i++)
	{
		token_address = TblGotoJump [i].token_address;
		address = TblLabelAddress [TblGotoJump [i].lbl_index];
		
		*(unsigned short *)(TokensFile_buffer+token_address) = address;
	}
}


//********************************************
//Vrifie si il ne manque pas un Endif...
//********************************************
BOOL ExprTokens_CheckStruct (void)
{
	if (n_IF_Jump) {error_code = ERROR_IF; return FALSE;}
	if (TotalDo) {error_code = ERROR_DO; return FALSE;}
	if (TotalRepeat) {error_code = ERROR_REPEAT; return FALSE;}
	if (TotalWhile) {error_code = ERROR_WHILE; return FALSE;}
	return TRUE;
}

//**********************************************************
//Ajoute une structure
//**********************************************************
BOOL ExprTokens_NewStruct (unsigned char x)
{
	if (nTblStruct>=MAX_STRUCT) {error_code = ERROR_STRUCT_OVERFLOW; return FALSE;}
	
	TblStruct [nTblStruct++] = x; 
	if (x!=STRUCT_IF) {
		if (nTblStructExitIf>=MAX_EXITIF) {error_code = ERROR_EXITIF_OVERFLOW; return FALSE;}
		nTblExitIf [nTblStructExitIf++] = 0;}
		
	return TRUE;
}

//**********************************************************
//Efface structure
//**********************************************************
void ExprTokens_DelStruct (unsigned short address)
{
	if (TblStruct [--nTblStruct]==STRUCT_IF) return;
	unsigned char nExitIf = nTblExitIf [nTblStructExitIf-1];
	
	for (int i=1; i<=nExitIf; i++)
		*(unsigned short *)(TokensFile_buffer+TblExitIf [TotalExitIf-i]) = address;
	
	TotalExitIf -= nTblExitIf [--nTblStructExitIf];
}